(*
Algorithm 1 of: 
  A Framework for Certified Self-Stabilization
  Case Study: Silent Self-Stabilizing k-Dominating Set on a Tree
https://hal.archives-ouvertes.fr/hal-01272158/document
*)

open Algo
open State
   

(* State predicates *)
  
let isRoot p = p.isRoot

let (isShort: State.t -> bool) =
  fun p -> p.alpha < k

let (isTall: State.t -> bool) =
  fun p -> p.alpha >= k

(* Actually unused *)
let (kDominator: 'v -> bool) = 
  fun p -> (p.alpha = k) || ((isShort p) && (isRoot p)) 


let (children: State.t -> State.t list -> State.t list) =
  fun p nl ->
  List.filter (fun q -> q.par = reply q) nl
    
let (shortChildren: State.t -> State.t list -> State.t list) =
  fun p nl ->
  let cl = List.filter (children p) nl in
  List.filter isShort cl

let (tallChildren: State.t -> State.t list -> State.t list) =
  fun p nl ->
  let cl = List.filter (children p) nl in
  List.filter isTall cl

let rec (max: State.t list -> int -> int) =
  fun sl cur ->
  match sl with
    [] -> cur
  | s::liste -> if (s.alpha) > cur then max liste (s.alpha) else max liste cur
  
let rec (min: State.t list -> int -> int) =
  fun sl cur ->
  match sl with
    [] -> cur
  | s::liste -> if (s.alpha) < cur then min liste (s.alpha) else min liste cur
  
let (maxAShort: State.t -> State.t list -> int) =
  fun p nl -> max (shortChildren p nl) (-1)

let (minATall: State.t -> State.t list -> int) =
  fun p nl -> min (tallChildren p nl) (2*k+1)

let (newAlpha: State.t -> State.t neighbor list -> int) =
  fun p nl ->
  let nl = List.map state nl in
  let mas = (maxAShort p nl) in
  let mit = (minATall p nl) in
  let res = if (mas + mit) <= (2*k - 2) then (mit + 1) else (mas + 1) in
  (*   Printf.printf "newAlpha -> %d\n%!" res; *)
  res
(*end macros*)

let (init_state: int -> string -> State.t) =
  fun _ pid ->
  (* assert(is_tree()); *)
  {
    isRoot = pid = "root"; (* ZZZ: The root of the tree should be named "root"! *)
    alpha = Random.int (2*k+1)
  }   
      
let (enable_f: State.t -> State.t neighbor list -> action list) =
  fun p nl -> if (p.alpha <> (newAlpha p nl)) then ["change_alpha"] else []

let (step_f : State.t -> State.t neighbor list -> action -> State.t ) =
  fun p nl a -> if a = "change_alpha" then {p with alpha = (newAlpha nl)} else assert false
