
(* Heuristique : pour aller vers le pire cas du Toy Example, on cherche à maximiser le nombre
  de propagations (action top-down) d'un résultat erroné. Pour un noeud que l'action top-down R
  n'est pas possible si l'action bottom-up S est possible, il faut donc faire S lorsque
  c'est nécessaire afin de continuer à avancer, mais le moins souvent possible. *)

open RdbgRun

let daemon_i = ref 0

type node_info = {
  lvl: int;
  s: bool;
  r: bool
}

(* Entrée : le contenu de !e.data
   Sortie : tuples ("n_S", activate) avec n le nom du noeud et S le nom de l'état, activate un booléen
   Attention, les noms ne doivent contenir que des lettres à cause du Data.type_to_string dans rdbgRun.ml:37 *)
let custom_daemon (data: Data.subst list): Data.subst list option =
  (* liste des noeuds-états *)
  let sl = List.filter (fun (n,v) -> String.length n > 5 && String.sub n 0 5 = "Enab_") data in

  (* regroupe les états par noeud *)
  let nodes = Hashtbl.create (List.length sl / 3) in
  List.iter (fun (n,v) ->
    let split = String.split_on_char '_' n in
    let enabled = match v with
    | Data.B b -> b
    | _ -> assert false
    in

    match split with
    | _::pid::state::[] -> (* Enab_pid_state *)
      let combine_info info =
        match state with
        | "S" -> {info with s=enabled}
        | "Rr" -> if info.lvl=0 then {info with r=enabled} else info
        | "Rp" -> if info.lvl>0 then {info with r=enabled} else info
        | _ -> assert false
      in
      (match Hashtbl.find_opt nodes pid with
      | None -> Hashtbl.add nodes pid (combine_info {lvl=Algo.level pid; s=false; r=false});
      | Some x -> Hashtbl.replace nodes pid (combine_info x);
      )
    | _ -> assert false
  ) sl;
  (* liste des états triés par niveau croissant (la racine en 1er) *)
  let nl = List.of_seq (Hashtbl.to_seq nodes) in
  let from_top = List.sort (fun (k1,v1) (k2,v2) -> Stdlib.compare v1.lvl v2.lvl) nl in

  (* Maximise le nombre de propagations en activant les noeuds tour à tour depuis la racine,
     à l'aide de la variable globale i *)
  let ni = !daemon_i in
  let n = List.length from_top in
  let res = List.flatten (List.mapi (fun i (k,v) ->
    let str_s = Printf.sprintf "%s_S" k in
    let str_rr = Printf.sprintf "%s_Rr" k in
    let str_rp = Printf.sprintf "%s_Rp" k in
    let s,rr,rp =
      if i = ni then
        (* c'est ce noeud-là qu'il faut activer, de préférence avec l'action R, sinon avec S puis R *)
        if v.r then (
          (* R maintenant *)
          daemon_i := (!daemon_i + 1) mod n;
          if v.lvl = 0 then
            false, true, false
          else
            false, false, true
        ) else (
          (* S au prochain coup *)
          assert v.s;
          true, false, false
        )
      else
        false, false, false
      in
      [(str_s, Data.B s); (str_rr, Data.B rr); (str_rp, Data.B rp)]
  ) from_top) in

  (* Simule le greedy daemon pour la fonction de potentiel du toy example : *)
  (* on fait R le + bas possible, sinon on fait S le + haut possible *)
  (* 
  let from_bottom = List.sort (fun (k1,v1) (k2,v2) -> Stdlib.compare v2.lvl v1.lvl) nl in
  let activate_r = List.find_opt (fun (_,info) -> info.r) from_bottom in
  let activate_s = List.find_opt (fun (_,info) -> info.s) from_top in

  (* résultat du démon *)
  let res = List.map (fun (n,v) ->
    let split = String.split_on_char '_' n in
    match split with
    | "Enab"::pid::state::[] ->
      let activate = g
        if v = Data.B true then
          if state <> "S" then
            match activate_r with
            | Some (node,info) -> node=pid
            | None -> false
          else if activate_r = None then
            match activate_s with
            | Some (node,info) -> node=pid
            | None -> false
          else
            false
        else
          false
      in
      let str = Printf.sprintf "%s_%s" pid state in
      (str, Data.B activate)
    | _ -> assert false
  ) sl
  in *)
  Some res

let is_end (e: RdbgEvent.t) =
  match (List.assoc_opt "potential" e.data) with
  | Some (Data.F 0.0) -> true
  | Some (Data.F x) -> false
  | _ -> assert false

let auto () =
  let stop = ref false in
  while not !stop do
    e := next_cond !e (fun e -> e.lang = "sasa");
    pe ();
    stop := is_end !e;
  done;
  Printf.printf "done in %d moves\n" (!e.step - 1);
  ()

let _ =
  rdbg_mv_hook := Some custom_daemon;
